'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''

import torch
import numpy as np
# import tensorflow as tf
from sklearn.metrics import classification_report
from nncodec.framework.applications.utils.metrics import get_topk_accuracy_per_batch
# from nncodec.framework.applications.models.tokenizer import Tokenizer
from nncodec.nnc_core import nnr_model
from contextlib import nullcontext
import re

DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

def evaluate_classification_model(model, criterion=None, testloader=None, testset=None,  min_sample_size=1000, max_batches=None,
                                  early_stopping_threshold=None, device=DEVICE, print_classification_report=False,
                                  return_predictions=False, verbose=False, rec_mdl=False):
    """
    Helper function to evaluate model on test dataset.

    Parameters
    ----------
    model: torch.nn.Module
        Neural network model.
    criterion: torch.nn.Criterion
        Criterion for loss calculation.
    testloader: torch.utils.data.DataLoader
        DataLoader that loaded testset.
    testset: torch.utils.data.dataset.Dataset
        Test dataset
    min_sample_size: int
        Minimum sample size used for early_stopping calculation. Default: 1000
    max_batches: int
        Maximum batches evaluated, by default evaluates the complete testset. Default: None
    early_stopping_threshold: int
        A value between 0-100 corresponding to the accuracy. If it drops under a given threshold
        the evaluation is stopped.
    device: str
        Device on which the model is evaluated: cpu or cuda.
    print_classification_report: bool
        If True print the complete confusion matrix for all the classes.
    return_predictions: bool
        If True return all the predictions for all samples, otherwise return the accuracy.
    verbose: bool
        If True print the progress bar of the evaluation.

    Return
    ------
    output: float | nd.array
        Accuracy or all predictions, depending on the given return_predictions parameter.
    """
    if isinstance(model, nnr_model.ModelExecute):
        criterion = model.handle.criterion
        testloader = model.test_loader
        testset = model.test_set
        device = model.device
        if rec_mdl:
            model = model.rec_model
        else:
            model = model.model

    if criterion is None:
        criterion = torch.nn.CrossEntropyLoss()

    model = model.to(device)
    model.eval()
    test_loss = []
    all_predictions = []
    all_labels = []
    correct = 0
    total = 0
    top5_acc = 0

    # set (verbose) iterator
    total_iterations = max_batches or len(testloader)
    # iterator = tqdm(enumerate(testloader), total=total_iterations, position=0, leave=True) if verbose else enumerate(testloader)
    iterator = enumerate(testloader)

    DeepLab_condition = model.__class__.__name__ == "DeepLabV3" if not (isinstance(model, torch.nn.DataParallel) or isinstance(model, torch.nn.parallel.DistributedDataParallel)) \
                            else model.module.__class__.__name__ == "DeepLabV3"

    if DeepLab_condition:
        from torchmetrics import JaccardIndex
        num_of_classes = model.classifier[len(model.classifier) - 1].weight.shape[0]
        jaccard = JaccardIndex(task="multiclass", num_classes=num_of_classes, average="macro").to(device)

    with torch.no_grad():
        for batch_idx, (inputs, targets) in iterator:
            inputs, targets = inputs.to(device), targets.to(device)
            outputs = model(inputs)

            if DeepLab_condition:
                outputs = outputs['out']
                targets = targets * (targets != 1) * 255
                targets = targets.squeeze(1).long()

            loss = criterion(outputs, targets)

            if outputs.size(1) > 5 and not DeepLab_condition:
                c1, c5 = get_topk_accuracy_per_batch(outputs, targets, topk=(1, 5))
                top5_acc += c5 * targets.size(0)

            test_loss.append(loss.item())
            _, predicted = outputs.max(1)
            total += targets.size(0) if not DeepLab_condition else targets.numel()
            correct += predicted.eq(targets).sum().item()
            all_predictions.append(np.array(predicted.cpu()))
            all_labels.append(np.array(targets.cpu()))

            acc = 100. * correct / total

            if DeepLab_condition:
                jaccard.update(predicted, targets)
                if verbose:
                    print('Running Test/Val mIOU (batch {}/{}) over all {} classes: {}'.format(batch_idx,
                                                                                               total_iterations,
                                                                                               num_of_classes,
                                                                                               jaccard.compute() * 100))

            if batch_idx == max_batches:
                break
            elif len(all_predictions) > min_sample_size and early_stopping_threshold is not None and \
                    acc < early_stopping_threshold:
                break

        acc = 100. * correct / total
        if top5_acc != 0:
            top5_acc = top5_acc / total

        if print_classification_report:
            print(classification_report(np.concatenate(all_labels), np.concatenate(all_predictions),
                                        target_names=list(testset.mapping.keys()),
                                        labels=list(testset.mapping.values())))

        if return_predictions:
            return np.concatenate(all_predictions)
        else:
            mean_test_loss = np.mean(test_loss)
            if DeepLab_condition:
                m_IoU = jaccard.compute() * 100
                return {'acc': acc, 'm_IoU': m_IoU, 'mean_test_loss': mean_test_loss}
            else:
                return {'acc': acc, 'top5_acc': float(top5_acc), 'mean_test_loss': mean_test_loss}

def evaluate_classification_model_TEF(model, test_loader, test_set, num_workers=8, verbose=0):

    _ , val_labels = zip(*test_set.imgs)

    y_pred = model.predict(test_loader, verbose=verbose, callbacks=None, max_queue_size=10, workers=num_workers,
                           use_multiprocessing=True)

    top5 = tf.keras.metrics.sparse_top_k_categorical_accuracy(val_labels, y_pred, k=5)
    top1 = tf.keras.metrics.sparse_categorical_accuracy(val_labels, y_pred)
    loss = tf.keras.metrics.sparse_categorical_crossentropy(val_labels, y_pred)

    acc = []
    acc.append((tf.keras.backend.sum(top1) / len(top1)).numpy() * 100)
    acc.append((tf.keras.backend.sum(top5) / len(top5)).numpy() * 100)
    acc.append((tf.keras.backend.mean(loss)).numpy())

    return acc

@torch.no_grad()
def evaluate_language_model(model, testloader, device='mps', max_batches=3, verbose=False, criterion=None, detokenize=False, args=None):

    if detokenize:
        import textwrap
        import json

        def is_number(s):
            try:
                float(s)  # float() can handle both integers and floats
                return True
            except ValueError:
                return False

        enc = Tokenizer(tokenizer_model=f"{args.tokenizer_path}")
        vocab_size = enc.sp_model.get_piece_size()
        vocabulary = [enc.sp_model.id_to_piece(i) for i in range(vocab_size)]

        global_predictions, global_gt, global_mse, global_rel_diff = {}, {}, {}, {}

    model = model.to(device)

    ptdtype = {"float32": torch.float32, "bfloat16": torch.bfloat16, "float16": torch.float16}["float16"]

    ctx = (
        nullcontext()
        if device == "cpu"
        else torch.amp.autocast(device_type=device.type, dtype=ptdtype)
    )

    if max_batches == None:
        try:
            max_batches = len(testloader)
        except:
            max_batches = testloader.dataset.num_samples

    model.eval()
    test_loss = []
    test_acc = []

    for k, batch in enumerate(testloader):

        if k == max_batches:
            break

        if detokenize:
            X, Y = batch, None
        else:
            X, Y = batch

        X = X.to(device, non_blocking=True)
        if Y is not None:
            Y = Y.to(device, non_blocking=True)
        with ctx:
            logits = model(X,Y)
            loss = model.last_loss

        if not detokenize:
            test_loss.append(loss.item())
            predictions = torch.argmax(logits, dim=-1)  # Shape: (batch_size, seq_length)
            correct = (predictions == Y).float()
            test_acc.append(correct.mean().item() * 100)

        if detokenize:
            output_sample = None
            text_width = 150

            exclude_strings = {'=', '-->', ''}

            def merge_consecutive_digits(token_list):  ## merging consecutiive digits to one number
                merged_digits = []
                i = 0
                while i < len(token_list):
                    if token_list[i] in {'', '='}:
                        merged_digits.append(token_list[i])
                        i += 1
                    elif re.match(r'^-?\d*\.?\d*$', token_list[i]) and token_list[i] != '':
                        num = token_list[i]
                        i += 1
                        while i < len(token_list) and re.match(r'^\d*\.?\d*$', token_list[i]):# and token_list[i] != '':
                            num += token_list[i]
                            i += 1
                        merged_digits.append(num)
                    else:
                        merged_digits.append(token_list[i])
                        i += 1
                return merged_digits

            inp_token = X[0,:]

            wrapped_text = [enc.decode(inp_token[i].int().tolist()) for i in range(X.shape[1])]

            arrow_indices = [i for i, s in enumerate(wrapped_text) if s == "-->"]
            if arrow_indices:
                input_sample = wrapped_text[:arrow_indices[-1] + 1]
                input_sample = merge_consecutive_digits(input_sample)
                print(f"Input:\n{textwrap.fill(' '.join(input_sample), width=text_width)} \n")

                output_sample = wrapped_text[arrow_indices[-1]:]
                model_input = inp_token[:arrow_indices[-1] + 1].unsqueeze(0)
                concat_results = []
                current_line_length = 0
                last_char = ""
                print(f"Predicting [...]\n")

                for i in range(max(args.max_seq_len, len(output_sample) + 1)):
                    if model_input.shape[1] > args.max_seq_len:
                        break
                    logits = model(model_input)

                    arg_max = torch.argmax(logits.squeeze(0), dim=1).item()
                    char = vocabulary[arg_max]

                    if char == 'time':
                        break
                    elif char == "<unk>":
                        break
                    elif char != '▁':
                        concat_results.append(char)
                        print(f'{char if ((is_number(last_char) or last_char in [".", "-"]) and (is_number(char) or char == ".")) else f" {char}"}', end='', flush=True)
                        current_line_length += len(char)
                        last_char = char
                    else:
                        if last_char and not (is_number(last_char) or last_char in [".", "-"]):
                            print(" ", end='', flush=True)
                        current_line_length += 1

                    model_input = torch.cat((model_input, torch.tensor([[arg_max]], device=device)), dim=1)

                    if current_line_length >= text_width:
                        print()
                        current_line_length = 0

            if output_sample:
                print("\n")
                print(f"Ground Truth:\n")
                output_sample = merge_consecutive_digits(output_sample)
                concat_results = merge_consecutive_digits(concat_results)
                print(f"{textwrap.fill(' '.join(output_sample), width=text_width)} \n")

                def extract_values(string_list):
                    values, i = {}, 0
                    while i < len(string_list) - 1:
                        key = string_list[i].strip()
                        if key not in exclude_strings:
                            j = i + 1
                            while j < len(string_list) and string_list[j].strip() in exclude_strings:
                                j += 1
                            if j < len(string_list):
                                try:
                                    value = float(string_list[j].strip())
                                    values[key] = value
                                except ValueError:
                                    pass
                        i += 1
                    return values

                predicted_values = extract_values(concat_results)
                ground_truth_values = extract_values(output_sample)

                rel_differences, diff = {}, {}
                for key in predicted_values:
                    if key in ground_truth_values:
                        diff[key] = predicted_values[key] - ground_truth_values[key]
                        if key in global_predictions:
                            global_predictions[key] += [predicted_values[key]]
                        else:
                            global_predictions[key] = [predicted_values[key]]
                        if key in global_gt:
                            global_gt[key] += [ground_truth_values[key]]
                        else:
                            global_gt[key] = [ground_truth_values[key]]

                for key in diff:
                    rel_differences[key] = (diff[key] / (ground_truth_values[key] + 1e-10) ) * 100
                    if key in global_rel_diff:
                        global_rel_diff[key] += [rel_differences[key]]
                    else:
                        global_rel_diff[key] = [rel_differences[key]]

                print("---------------------------------------------------------------------------------------\n")
                print(f"Relative differences wrt. ground truth: ")
                [print(f"{rel_diff}: {rel_differences[rel_diff]:.1f}%") for rel_diff in rel_differences]
                print("---------------------------------------------------------------------------------------\n")
                print(f"Running absolute mean relative differences wrt. ground truth (global):")
                [print(f"{grel_diff}: {np.mean(np.abs(np.array(global_rel_diff[grel_diff]))):.1f}%") for grel_diff in global_rel_diff]
                print("---------------------------------------------------------------------------------------\n")


                with open(f'{args.results}/test_predictions.json', 'w') as f:
                    json.dump(global_predictions, f)
                with open(f'{args.results}/test_ground_truth.json', 'w') as f:
                    json.dump(global_gt, f)
                with open(f'{args.results}/test_rel_differences.json', 'w') as f:
                    json.dump(global_rel_diff, f)

    loss = torch.mean(torch.tensor(test_loss)).item()
    acc = torch.mean(torch.tensor(test_acc)).item()
    ppl = torch.exp(torch.tensor(loss)).item()
    model.train()
    print(f'top1-acc: {acc:.3f}%, ppl: {ppl:.3f}, loss: {loss:.3f}')
    return {'top1_acc': acc, 'ppl': ppl, 'loss': loss}



'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''
    
import sys
assert sys.version_info >= (3, 6)

import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import numpy as np
import copy
from timeit import default_timer as timer
from nncodec import nnc_core
from nncodec.nnc_core import nnr_model
from nncodec.framework import tensorflow_model, pytorch_model, use_case_init


def __print_output_line( outputString, verbose=True ):
    if verbose:
        sys.stdout.write(outputString)
        sys.stdout.flush()
        
def guess_block_id_and_param_type(model_struct, add_lsa_params=False):
    if tensorflow_model.is_tef_model(model_struct):
        nnc_mdl, _, model_parameters = tensorflow_model.create_NNC_model_instance_from_object(
                 model_struct,
                )
        block_id_and_param_type = nnc_mdl.guess_block_id_and_param_type(model_parameters)
    elif pytorch_model.is_pyt_model(model_struct):
        nnc_mdl, _, model_parameters = pytorch_model.create_NNC_model_instance_from_object(
                 model_struct,
                )
        block_id_and_param_type = nnc_mdl.guess_block_id_and_param_type(model_parameters)
    else:
        print("INFO: guess_block_id_and_param_type is only applicable to Pytorch and Tensorflow models! block_id_and_param_type has been set to 'None'")
        block_id_and_param_type=None
        
    if block_id_and_param_type and add_lsa_params:
        for param, parType in block_id_and_param_type["parameter_type"].items():
            if parType == "weight":
                lsa_param = param + "_scaling"
                block_id_and_param_type["parameter_type"][lsa_param] = "weight.ls"
                block_id_and_param_type["block_identifier"][lsa_param] = block_id_and_param_type["block_identifier"][param]
    
    if block_id_and_param_type:    
        blkIdParamTypeOk = nnc_core.nnr_model.sanity_check_block_id_and_param_type( block_id_and_param_type, model_parameters )
        if blkIdParamTypeOk == False:
            print("INFO: Sanity check for block_id_and_param_type failed! block_id_and_param_type has been set to 'None'!")
            block_id_and_param_type = None
    
    return block_id_and_param_type
        

def compress_model( model_path_or_object,
                    bitstream_path="./bitstream.nnc",
                    qp=-38,
                    qp_density=2,
                    nonweight_qp=-75,
                    qp_per_tensor=None,
                    use_dq=True,
                    codebook_mode=0,
                    scan_order=0,
                    lambda_scale=0,
                    param_opt=True,
                    cabac_unary_length_minus1=10,
                    opt_qp=False,
                    ioq=False,
                    bnf=False,
                    lsa=False,
                    fine_tune=False,
                    row_skipping=False,
                    tca=False,
                    block_id_and_param_type=None,
                    model_name=None,
                    model_executer=None,
                    model_struct=None,
                    dataset_path=None,
                    use_case=None,
                    learning_rate=1e-4,
                    batch_size=64,
                    epochs=30,
                    max_batches=None,
                    num_workers=8,
                    return_model_data=False,
                    verbose=True,
                    return_bitstream=False,
                    wandb_logging=False,
                    approx_param_base = None,
                    device_id = 0,
                    int_quant_bw = False,
                   ):

    is_pyt_model = False
    is_tef_model = False
    dataset_path = None if dataset_path is None else os.path.expanduser(dataset_path)
        
    if tensorflow_model.is_tef_model(model_path_or_object): 
        if bnf:
            print("WARNING: Batch-norm folding (BNF) requires the tensors to be shaped such that the first dimensions corresponds to the number of output channels, which is usually not the case for TensorFlow. For further details refer to the Wiki!")
        if lsa:
            print("INFO: LSA not yet supported for TensorFlow models. 'lsa' has been set to false!")
            lsa = False
        is_tef_model = True
        if model_executer:
                nnc_mdl, _, model_parameters = tensorflow_model.create_NNC_model_instance_from_object(
                 model_path_or_object,
                )
        else:
            if use_case != None:
                assert use_case in use_case_init.use_cases.keys(), \
                    f"use case must be one of {list(use_case_init.use_cases.keys())}"
            nnc_mdl, nnc_mdl_executer, model_parameters = tensorflow_model.create_NNC_model_instance_from_object(
                model_path_or_object,
                dataset_path=dataset_path,
                lr=learning_rate,
                batch_size=batch_size,
                num_workers=num_workers,
                model_struct=model_struct,
                model_name=model_name,
                use_case=use_case
                )
    elif pytorch_model.is_pyt_model(model_path_or_object):
        is_pyt_model = True
        if model_executer:
                nnc_mdl, _, model_parameters = pytorch_model.create_NNC_model_instance_from_object(
                 model_path_or_object,
                )
        else:
            if use_case != None:
                assert use_case in use_case_init.use_cases.keys(), \
                    f"use case must be one of {list(use_case_init.use_cases.keys())}"
            nnc_mdl, nnc_mdl_executer, model_parameters = pytorch_model.create_NNC_model_instance_from_object(
                model_path_or_object,
                dataset_path=dataset_path,
                lr=learning_rate,
                batch_size=batch_size,
                num_workers=num_workers,
                model_struct=model_struct,
                lsa=lsa,
                epochs=epochs,
                max_batches=max_batches,
                use_case=use_case
                )
    elif os.path.exists( os.path.expanduser(model_path_or_object)):
        model_path_or_object = os.path.expanduser(model_path_or_object)
        if model_path_or_object.endswith(".h5") or model_path_or_object.endswith(".hdf5") or model_path_or_object.endswith(".tf"):
            if bnf:
                print("WARNING: Batch-norm folding (BNF) requires the tensors to be shaped such that the first dimensions corresponds to the number of output channels, which is usually not the case for TensorFlow. For further details refer to the Wiki!")
            if lsa:
                print("INFO: LSA not yet supported for TensorFlow models. 'lsa' has been set to false!")
                lsa = False
            is_tef_model = True
            if model_executer:
                nnc_mdl, _, model_parameters = tensorflow_model.create_NNC_model_instance_from_file(
                 model_path_or_object,
                )
            else:
                nnc_mdl, nnc_mdl_executer, model_parameters = tensorflow_model.create_NNC_model_instance_from_file(
                    model_path_or_object,
                    dataset_path=dataset_path,
                    lr=learning_rate,
                    batch_size=batch_size,
                    num_workers=num_workers,
                    model_struct=model_struct,
                    model_name=model_name
                    )

        elif model_path_or_object.endswith(".pt") or model_path_or_object.endswith(".pth"):
            is_pyt_model = True
            if model_executer:
                nnc_mdl, _, model_parameters = pytorch_model.create_NNC_model_instance_from_file(
                 model_path_or_object,
                )
            else:    
                nnc_mdl, nnc_mdl_executer, model_parameters = pytorch_model.create_NNC_model_instance_from_file(
                    model_path_or_object,
                    dataset_path=dataset_path,
                    lr=learning_rate,
                    batch_size=batch_size,
                    num_workers=num_workers,
                    model_struct=model_struct,
                    lsa=lsa,
                    epochs=epochs,
                    max_batches=max_batches,
                    )

        else:
            nnc_mdl, model_parameters = nnr_model.create_NNC_model_instance_from_file( model_path_or_object )
            nnc_mdl_executer = None

    else:
        raise SystemExit("Can't find path or object {}".format(model_path_or_object))

    if model_executer:
        nnc_mdl_executer = model_executer    
    
    if block_id_and_param_type is None and (bnf or lsa) and (is_pyt_model or is_tef_model):
        if is_pyt_model: #ADDED for ICML
            bn_info = nnc_mdl.get_torch_bn_info(model_path_or_object)
            block_id_and_param_type = nnc_mdl.guess_block_id_and_param_type(model_parameters, bn_info=bn_info)
        else:
            block_id_and_param_type = nnc_mdl.guess_block_id_and_param_type(model_parameters)
        blkIdParamTypeOk = nnc_core.nnr_model.sanity_check_block_id_and_param_type( block_id_and_param_type, model_parameters )
        if blkIdParamTypeOk == False:
            print("INFO: Sanity check for block_id_and_param_type failed! block_id_and_param_type has been set to 'None', and the flags 'lsa' and 'bnf' have been set to 'False'!")
            block_id_and_param_type = None
            lsa = False
            bnf = False
            if model_executer:
                model_executer.model = model_executer.original_model
                del model_executer.original_model
            for key in model_parameters.keys():
                if "weight_scaling" in key:
                    del model_parameters[key]
            

    bitstream = compress(   model_parameters,
                            bitstream_path=bitstream_path,
                            qp=qp,
                            qp_density=qp_density,
                            nonweight_qp=nonweight_qp,
                            qp_per_tensor=qp_per_tensor,
                            use_dq=use_dq,
                            codebook_mode=codebook_mode,
                            scan_order=scan_order,
                            lambda_scale=lambda_scale,
                            param_opt=param_opt,
                            cabac_unary_length_minus1=cabac_unary_length_minus1,
                            opt_qp=opt_qp,
                            ioq=ioq,
                            bnf=bnf,
                            lsa=lsa,
                            fine_tune=fine_tune,
                            row_skipping=row_skipping,
                            tca=tca,
                            block_id_and_param_type=block_id_and_param_type,
                            model=nnc_mdl,
                            model_executer=nnc_mdl_executer,
                            verbose=verbose,
                            return_bitstream=return_bitstream,
                            wandb_logging=wandb_logging,
                            approx_param_base=approx_param_base,
                            device_id=device_id,
                            int_quant_bw = int_quant_bw,
                            )

    # if bnf: #ADDED for ICML
    #     block_id_and_param_type["bnf_matching"], bitstream = bitstream[1], bitstream[0]

    if return_model_data==True:# or (bnf and is_pyt_model):
        if return_bitstream:
            return bitstream, block_id_and_param_type
        else:
            return block_id_and_param_type
    elif return_bitstream:
        return bitstream


def compress( 
    parameter_dict,
    bitstream_path="./bitstream.nnc",
    qp=-38,
    qp_density=2,
    nonweight_qp=-75,
    qp_per_tensor=None,
    use_dq=True,
    codebook_mode=0,
    scan_order=0,
    lambda_scale=0,
    param_opt=True,
    cabac_unary_length_minus1=10,
    opt_qp=False,
    ioq=False,
    bnf=False,
    lsa=False,
    fine_tune=False,
    row_skipping=False,
    tca=False,
    hdsp_tool=None,
    pre_signalling=False,
    block_id_and_param_type=None,
    model=None,
    model_executer=None,
    verbose=False,
    return_bitstream=False,
    bnf_mapping=False,
    wandb_logging=False,
    approx_param_base=None,
    update_base_param=False,
    device_id=0,
    compress_differences=False,
    int_quant_bw=False,
    quantize_only=False
    ):

    try:
        start = timer()
        start_overall = start
        __print_output_line("INITIALIZE APPROXIMATOR AND ENCODER...", verbose=verbose)
        if isinstance(parameter_dict, dict) and all( [isinstance(a, np.ndarray) for a in parameter_dict.values()] ) and (all([ (a.dtype==np.float32 or a.dtype==np.int32) for a in parameter_dict.values()])):
            model_parameters = parameter_dict

            if isinstance(model, nnc_core.nnr_model.NNRModel):
                nnc_mdl = model
            else:
                nnc_mdl = nnc_core.nnr_model.NNRModel(parameter_dict)

            if model_executer is not None:
                assert isinstance( model_executer, nnc_core.nnr_model.ModelExecute ), "model_executer must be of type ModelExecute!"
        else:
            raise SystemExit("Parameter dict must be a dict (key-value pairs). The keys shall be stings, specifying the tensor names. The values shalls be numpy arrays (ndarray) of type float32 or int32!")
    except:
        raise SystemExit("Can not read parameter_dict: {}".format(parameter_dict))

    if block_id_and_param_type is not None:
        blkIdParamTypeOk = nnc_core.nnr_model.sanity_check_block_id_and_param_type( block_id_and_param_type, parameter_dict )
        if blkIdParamTypeOk:
            nnc_core.nnr_model.set_block_id_and_param_type( nnc_mdl.model_info , block_id_and_param_type )
        else:
            print("INFO: Sanity check for block_id_and_param_type failed! block_id_and_param_type has been set to 'None', and the flags 'lsa' and 'bnf' have been set to 'False'!")
            block_id_and_param_type = None
            lsa = False
            bnf = False


    if model_executer:
        if lsa and not model_executer.has_tune_lsa():
            print("INFO: Tuning (training) of LSA parameters (tune_model) not implemented by model_executer! 'lsa' has been set to 'False'!")
            lsa = False
        if fine_tune and not model_executer.has_tune_ft():
            print("INFO: Fine tuning (training) of parameters (tune_model) not implemented by model_executer! 'fine_tune' has been set to 'False'!")
            fine_tune = False
        if ioq and not model_executer.has_eval():
            print("INFO: Evaluation (inference on a reduced dataset) of parameters (eval_model) not implemented by model_executer! ioq' has been set to 'False'!")
            ioq = False

    ##INITIALIZATION
    approx_data =  nnc_core.approximator.init_approx_data(  model_parameters,
                                                            nnc_mdl.model_info,
                                                            qp_density=qp_density,
                                                            scan_order=scan_order
                                                         )

    ApproxInfoO = nnc_core.approximator.ApproxInfo( approx_data,
                                                    nnc_mdl.model_info,
                                                    "uniform" if codebook_mode==0 else "codebook",
                                                    codebook_mode,
                                                    qp,
                                                    opt_qp,
                                                    not use_dq,
                                                    cabac_unary_length_minus1,
                                                    lambda_scale,
                                                    nonweight_qp=nonweight_qp,
                                                    qp_per_tensor=qp_per_tensor,
                                                    int_quant_bw=int_quant_bw
                                                )
    approx_info = ApproxInfoO.approx_info

    enc_info = {
            "cabac_unary_length_minus1" : cabac_unary_length_minus1,
            "param_opt_flag"     : param_opt,
            "partial_data_counter": 0,  # TODO parameterize
            "general_profile_idc": 1,  # TODO parameterize
        }

    if enc_info.get("general_profile_idc", 0):
        """
        If mps_parent_signalling_enabled_flag is equal to 0, parent_node_id_present_flag is not present for any NDU 
        (i.e. the bitstream represents a base neural network). If mps_parent_signalling_enabled_flag is equal to 1, 
        the bitstream represents a differential update of a base neural network.
        
        parent_node_id_present_flag indicates whether the NDU represents a differential update of a base neural network. 
        It shall be set to 1 for NDUs representing differential updates, and to 0 otherwise. A value
        """
        enc_info["node_id_present_flag"] = 1
        enc_info["device_id"] = device_id
        enc_info["mps_parent_signalling_enabled_flag"] = 1 if compress_differences or tca else 0
        enc_info["parent_node_id_present_flag"] = 1 if tca else 0
        if enc_info["parent_node_id_present_flag"]:
            enc_info["parent_node_id_type"] = nnc_core.hls.ParentNodeIdType.ICNN_NDU_ID
            enc_info["temporal_context_modeling_flag"] = 1 if tca else 0
            enc_info["parent_device_id"] = 0

        enc_info["row_skip_enabled_flag"] = 1 if row_skipping else 0
        enc_info["pre_signalling"] = 1 if pre_signalling else 0
        # enc_info["nnr_pt_block_enabled_flag"] = nnr_pt_block_enabled_flag
    else:
        enc_info["mps_parent_signalling_enabled_flag"] = 0
        enc_info["parent_node_id_present_flag"] = 0
        enc_info["node_id_present_flag"] = 0

    end = timer()
    __print_output_line("DONE in {:.4f} s\n".format(end-start), verbose=verbose)

    ##PREPROCESSING
    if ioq and not bnf_mapping:
        assert model_executer is not None, "model_executer must be available in order to run IOQ!"
        start = timer()
        __print_output_line("PREPROCESSING, IOQ...\n", verbose=verbose)
        nnc_core.approximator.inference_based_qp_opt(
            approx_info,
            nnc_mdl.model_info,
            model_executer,
            approx_data,
            enc_info,
            verbose=verbose,
        )
        end = timer()
        __print_output_line("DONE in {:.4f} s\n".format( end-start ), verbose=verbose)

    ##LSA and FT
    if (lsa or fine_tune) and not bnf_mapping:
        assert model_executer is not None, "model_executer must be available in order to run LSA and/or FT!"
        start = timer()
        __print_output_line("PREPROCESSING, LSA/FT...\n", verbose=verbose)
        nnc_core.approximator.run_ft_and_lsa(
            nnc_mdl.model_info,
            approx_data,
            ApproxInfoO,
            model_executer,
            block_id_and_param_type,
            lsa,
            fine_tune,
            use_dq,
            verbose,
            wandb_logging
        )
        end = timer()
        __print_output_line("DONE in {:.4f} s\n".format( end-start ), verbose=verbose)
    ##BNF
    if bnf or bnf_mapping:
        start = timer()
        __print_output_line("PREPROCESSING, BNF...", verbose=verbose)
        nnc_core.approximator.fold_bn(nnc_mdl.model_info, approx_data, ApproxInfoO, bnf_mapping=bnf_mapping)
        end = timer()
        __print_output_line("DONE in {:.4f} s\n".format(end-start), verbose=verbose)
        if bnf_mapping:
            return nnc_mdl.model_info

    #####QUANTIZATION AND ENCODING
    start = timer()
    __print_output_line("APPROXIMATING WITH METHOD {}...".format(approx_info["approx_method"]), verbose=verbose)
    approx_data_enc = nnc_core.approximator.approx( approx_info,
                                                nnc_mdl.model_info,
                                                approx_data,
                                                enc_info
                                               )
    end = timer()
    __print_output_line("DONE in {:.4f} s\n".format( end-start ), verbose=verbose)

    if quantize_only:
        nnc_core.approximator.rec(approx_data_enc)
        return approx_data_enc["parameters"]

    start = timer()
    __print_output_line("ENCODING...", verbose=verbose)
    bitstream, _ = nnc_core.coder.encode(enc_info=enc_info,
                                         model_info=nnc_mdl.model_info,
                                         approx_data=approx_data_enc,
                                         approx_param_base=approx_param_base,
                                         update_base_param=update_base_param,
                                         tool_if=hdsp_tool
                                         )
    end = timer()
    __print_output_line("DONE in {:.4f} s\n".format( end-start ), verbose=verbose)

    original_size = nnc_mdl.model_info["original_size"]

    __print_output_line("COMPRESSED FROM {} BYTES TO {} BYTES ({:.2f} KB, {:.2f} MB, COMPRESSION RATIO: {:.2f} %) in {:.4f} s\n".format(original_size, len(bitstream), len(bitstream)/1000.0, len(bitstream)/1000000.0, len(bitstream)/original_size*100, end-start_overall), verbose=verbose)

    if hdsp_tool is not None:
        hdsp_tool.add_data_to_hist(approx_data_enc['parameters'])

    if bitstream_path is not None:
        with open( bitstream_path, "wb" ) as br_file:
            br_file.write( bitstream )

    if return_bitstream:
        return bitstream


def decompress( bitstream_or_path,
                block_id_and_param_type=None,
                return_model_information=False,
                return_hls=False,
                verbose=False,
                reconstruct_lsa=False,
                reconstruct_bnf=False,
                approx_param_base=None,
                update_base_param=False,
                hdsp_tool=None,
                internal_states_path=None,
                ):

    dec_model_info  = {'parameter_type': {},
                      'parameter_dimensions': {},
                      'parameter_index': {},
                      'block_identifier': {},
                      'topology_storage_format' : None,
                      'topology_compression_format' : None,
                      'performance_maps' : { "mps" : {}, "lps" : {}},
                      'performance_map_flags' : { "mps_sparsification_flag" : {}, "lps_sparsification_flag" : {},
                                                  "mps_pruning_flag" : {}, "lps_pruning_flag" : {},
                                                  "mps_unification_flag" : {}, "lps_unification_flag" : {},
                                                  "mps_decomposition_performance_map_flag" : {}, "lps_decomposition_performance_map_flag" : {},
                                                }
                      }

    model_information = { 'topology_storage_format' : None,
                          'performance_maps' : {},
                          'performance_map_flags' : {}
                        }

    if block_id_and_param_type is not None:
        blkIdParamTypeOk = nnc_core.nnr_model.sanity_check_block_id_and_param_type( block_id_and_param_type )
        if blkIdParamTypeOk == False:
            print("INFO: Sanity check for block_id_and_param_type failed! block_id_and_param_type has been set to 'None'!")
            block_id_and_param_type = None
        else:
            nnc_core.nnr_model.set_block_id_and_param_type( dec_model_info, block_id_and_param_type )

    hls_bytes = {}
    oob_dict = {}
    start = timer()
    __print_output_line("DECODING...", verbose=verbose)
    if isinstance(bitstream_or_path, bytearray):
        bitstream = bitstream_or_path
    elif os.path.exists(os.path.expanduser(bitstream_or_path)):
        with open( os.path.expanduser(bitstream_or_path), "rb" ) as br_file:
            bitstream = br_file.read()
    else:
        raise SystemExit( "Could not read bitstream or bitstream_path: {}".format(bitstream_or_path) )

    if internal_states_path and approx_param_base is None: # loading co-located params for temporal tool
        ndu_header = nnc_core.coder.decode_ndu_unit_header(copy.deepcopy(bitstream), dec_model_info, hls_stats=hls_bytes)
        _int_states_path = internal_states_path + f"/client_ID{ndu_header['device_id']}_internal_states.npz"
        loaded_states = np.load(_int_states_path, allow_pickle=True)  # TODO get rid of allow_pickle
        loaded_internal_states = {k: loaded_states[k].item() for k in loaded_states.files}
        approx_param_base = loaded_internal_states['approx_param_base']

    dec_approx_data = nnc_core.coder.decode(bitstream, dec_model_info, hls_stats=hls_bytes, oob_dict=oob_dict,
                                            approx_param_base=approx_param_base, update_base_param=update_base_param,
                                            return_hls=return_hls, tool_if=hdsp_tool)

    if hdsp_tool is not None and return_hls and dec_approx_data["dec_hls"]["mps"]["mps_parent_signalling_enabled_flag"]:
        hdsp_tool.add_data_to_hist(dec_approx_data["parameters"])

    if internal_states_path and approx_param_base["parameters"]:
        np.savez(f"{_int_states_path}", **loaded_internal_states)

    end = timer()
    __print_output_line("DONE in {:.4f} s\n".format( end-start ), verbose=verbose)

    start = timer()
    rec_approx_data = dec_approx_data
    __print_output_line("RECONSTRUCTING...", verbose=verbose)
    nnc_core.approximator.rec(rec_approx_data )
    if reconstruct_bnf: ## TODO: check if there are cases where must be dis/enabled
        nnc_core.approximator.unfold_bn(dec_model_info, rec_approx_data)
    if reconstruct_lsa: ## TODO: check if there are cases where must be dis/enabled
        nnc_core.approximator.apply_lsa(dec_model_info, rec_approx_data)
    rec_approx_data = nnc_core.approximator.recompose_params( dec_model_info, rec_approx_data)
    end = timer()
    __print_output_line("DONE in {:.4f} s\n".format( end-start ), verbose=verbose)
    
    if return_model_information:
        model_information["topology_storage_format"] = dec_model_info["topology_storage_format"]
        model_information["performance_maps"]        = dec_model_info["performance_maps"]
        model_information["performance_map_flags"]   = dec_model_info["performance_map_flags"]

        return rec_approx_data["parameters"], model_information
    elif return_hls:
        return (rec_approx_data["parameters"], dec_approx_data["dec_hls"])
    else:
        return rec_approx_data["parameters"]


def decompress_model( bitstream_or_path,
                      model_path=None,#"./rec.mdl",
                      block_id_and_param_type=None,
                      model_struct=None,
                      model_executer=None,
                      model_name=None, 
                      dataset_path=None, 
                      batch_size=64, 
                      num_workers=8,
                      reconstruct_bnf=True,
                      reconstruct_lsa=True,
                      test_model=False,
                      return_model_information=False,
                      return_decompressed_model=False,
                      verbose=True,
                    ):
    
    if block_id_and_param_type is not None:
        blkIdParamTypeOk = nnc_core.nnr_model.sanity_check_block_id_and_param_type( block_id_and_param_type )
        if blkIdParamTypeOk == False:
            print("INFO: Sanity check for block_id_and_param_type failed! block_id_and_param_type has been set to 'None'!")
            block_id_and_param_type = None
    
    model_dict, model_information = decompress(bitstream_or_path, 
                                        block_id_and_param_type=block_id_and_param_type, 
                                        return_model_information=True,
                                        reconstruct_lsa=reconstruct_lsa,
                                        reconstruct_bnf=reconstruct_bnf
                                       )

    model_with_decoded_parameters = None

    if model_information["topology_storage_format"] == nnc_core.nnr_model.TopologyStorageFormat.NNR_TPL_PYT:
        if model_path == None:
            model_path="./rec.pt"

        pytorch_model.save_to_pytorch_file( model_dict, model_path )

        if ( (model_struct and dataset_path) or model_executer ) and test_model:
            
            if model_executer:
                nnc_mdl_executer = model_executer
            else:    
                _, nnc_mdl_executer, _ = pytorch_model.create_NNC_model_instance_from_file(
                    model_path,
                    dataset_path=dataset_path,
                    batch_size=batch_size,
                    num_workers=num_workers,
                    model_struct=model_struct,
                    )
            
            acc = nnc_mdl_executer.test_model(
                    model_dict,
                    verbose=verbose
                    )
            print(acc)
        
        if model_struct and return_decompressed_model:
            model_with_decoded_parameters = pytorch_model.get_model_file_with_parameters(model_struct=model_struct, parameters=model_dict)
                
    elif model_information["topology_storage_format"] == nnc_core.nnr_model.TopologyStorageFormat.NNR_TPL_TEF:
        if model_path == None:
            model_path="./rec.h5"

        tensorflow_model.save_to_tensorflow_file( model_dict, model_path )
        
        if ( (model_struct and dataset_path) or model_executer ) and test_model:
            
            if model_executer:
                nnc_mdl_executer = model_executer
            else:
                _, nnc_mdl_executer, _ = tensorflow_model.create_NNC_model_instance_from_file(
                    model_path,
                    dataset_path=dataset_path,
                    batch_size=batch_size,
                    num_workers=num_workers,
                    model_struct=model_struct,
                    model_name=model_name 
                    )

            acc = nnc_mdl_executer.test_model(
                    model_dict,
                    verbose=verbose
                    )
            print(acc)

        if model_struct and return_decompressed_model:
           model_with_decoded_parameters = tensorflow_model.get_model_file_with_parameters(model_struct=model_struct, parameters=model_dict)

    elif model_information["topology_storage_format"] == nnc_core.nnr_model.TopologyStorageFormat.NNR_TPL_UNREC or model_information["topology_storage_format"] == None:
        if model_path == None:
            model_path="./rec.mdl"

        nnr_model.save_to_pickled_file( model_dict, model_path )

        if model_executer and test_model:
            nnc_mdl_executer = model_executer
            acc = nnc_mdl_executer.test_model(
                model_dict,
                verbose=verbose
            )
            print(acc)

    else:
        raise SystemExit( "Topology Storage Format not yet supported: {}".format( model_information["topology_storage_format"] ) )
    
    if return_decompressed_model and return_model_information:
        return model_with_decoded_parameters, model_information
    elif return_decompressed_model:
        return model_with_decoded_parameters
    elif return_model_information:
        return model_information


'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or
other Intellectual Property Rights other than the copyrights concerning
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

References:

    1)
    https://github.com/akamaster/pytorch_resnet_cifar10
    Proper ResNet Implementation for CIFAR10/CIFAR100 in PyTorch
    BSD 2-Clause "Simplified" License
    Copyright (c) 2018, Yerlan Idelbayev

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
    following conditions are met:
        1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
        following disclaimer.
        2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
        following disclaimer in the documentation and/or other materials provided with the distribution.
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
    INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
    WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
    USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    2)
    https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py

    3)
    Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun:
    Deep Residual Learning for Image Recognition. https://arxiv.org/abs/1512.03385
'''

import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init

def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )

def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)

class BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out

class Bottleneck(nn.Module):
    expansion = 4

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out

class SplitResNet(nn.Module):
    def __init__(
        self,
        block,
        layers,
        num_classes=10,
        zero_init_residual=False,
        groups=1,
        width_per_group=64,
        replace_stride_with_dilation=None,
        norm_layer=None,
        split=None
    ):
        super(SplitResNet, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer
        self.split = split

        self.inplanes = 16
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be None "
                "or a 3-element tuple, got {}".format(replace_stride_with_dilation)
            )

        self.groups = groups
        self.base_width = width_per_group

        if split == 'client' or split == None:
            self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=3, stride=1, padding=1, bias=False)
            self.bn1 = nn.BatchNorm2d(self.inplanes)
            self.relu = nn.ReLU(inplace=True)

        self.layer1 = self._make_layer(block, 16, layers[0])

        if split == 'server' or split == None:
            self.layer2 = self._make_layer(block, 32, layers[1], stride=2)
            self.layer3 = self._make_layer(block, 64, layers[2], stride=2)
            self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
            self.fc = nn.Linear(64 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or (self.split != 'server' and self.inplanes != planes * block.expansion):
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        if self.split == 'server' and stride == 1:
            inplanes = planes * block.expansion
        else:
            inplanes = self.inplanes

        layers = []
        layers.append(
            block(
                inplanes,
                planes,
                stride,
                downsample=downsample,
                groups=self.groups,
                base_width=self.base_width,
                dilation=previous_dilation,
                norm_layer=norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def forward(self, x):
        if self.split == 'client':
            x = self.conv1(x)
            x = self.bn1(x)
            x = self.relu(x)
            x = self.layer1(x)
        elif self.split == 'server':
            x = self.layer1(x)
            x = self.layer2(x)
            x = self.layer3(x)
            x = self.avgpool(x)
            x_f = x.view(x.size(0), -1)
            x = self.fc(x_f)
        else:
            x = self.conv1(x)
            x = self.bn1(x)
            x = self.relu(x)  # B x 16 x 32 x 32
            x = self.layer1(x)  # B x 16 x 32 x 32
            x = self.layer2(x)  # B x 32 x 16 x 16
            x = self.layer3(x)  # B x 64 x 8 x 8
            x = self.avgpool(x)  # B x 64 x 1 x 1
            x_f = x.view(x.size(0), -1)  # B x 64
            x = self.fc(x_f)  # B x num_classes
        return x

def resnet56_client_split(**kwargs):
    return SplitResNet(Bottleneck, [1], split='client', **kwargs)

def resnet56_server_split(class_num, **kwargs):
    return SplitResNet(Bottleneck, [5, 6, 6], class_num, split='server', **kwargs)

def resnet56(class_num, **kwargs):
    return SplitResNet(Bottleneck, [6, 6, 6], class_num, **kwargs)


def _weights_init(m):
    if isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d):
        init.kaiming_normal_(m.weight)
        # init.xavier_normal_(m.weight)

class LambdaLayer(nn.Module):
    def __init__(self, lambd):
        super(LambdaLayer, self).__init__()
        self.lambd = lambd

    def forward(self, x):
        return self.lambd(x)

class BasicBlockcifar(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlockcifar, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != planes:
            self.shortcut = LambdaLayer(lambda x:
                                        F.pad(x[:, :, ::2, ::2], (0, 0, 0, 0, planes // 4, planes // 4), "constant", 0))

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet20(nn.Module):
    def __init__(self, block, num_blocks, num_classes=10):
        super(ResNet20, self).__init__()
        self.in_planes = 16

        self.conv1 = nn.Conv2d(3, 16, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.layer1 = self._make_layer(block, 16, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 32, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 64, num_blocks[2], stride=2)
        self.linear = nn.Linear(64, num_classes)

        self.apply(_weights_init)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion

        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = F.avg_pool2d(out, out.size()[3])
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out

class ResNet20_server(nn.Module):
    def __init__(self, block, num_blocks, num_classes=10):
        super(ResNet20_server, self).__init__()
        self.in_planes = 16
        self.smashed_data = None
        self.layer1 = self._make_layer(block, 16, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 32, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 64, num_blocks[2], stride=2)
        self.linear = nn.Linear(64, num_classes)
        self.apply(_weights_init)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def _save_smashed_data(self, module, grad_input, grad_output):
        self.smashed_data = grad_input[0]

    def forward(self, x):
        out = self.layer1(x)
        out = self.layer2(out)
        out = self.layer3(out)
        out = F.avg_pool2d(out, out.size()[3])
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out

class ResNet20_client(nn.Module):
    def __init__(self, block):
        super(ResNet20_client, self).__init__()
        self.in_planes = 16
        self.conv1 = nn.Conv2d(3, 16, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.layer1 = nn.Sequential(block(self.in_planes, 16))
        self.apply(_weights_init)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        return self.layer1(out)

def resnet20():
    return ResNet20(BasicBlockcifar, [3, 3, 3])

def resnet20_client_split():
    return ResNet20_client(BasicBlockcifar)

def resnet20_server_split():
    return ResNet20_server(BasicBlockcifar, [2, 3, 3])



'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''

"""
Defines data and model transforms.
"""

import torch
import copy
import torch.nn as nn
from torch.functional import F
import cv2 as cv
import numpy as np
from torchvision import transforms
from torch.utils.data import random_split, DataLoader

MDL_TRAFOS = [
    "model_transform_ImageNet_to_CIFAR100",
    "LSA"
]

DATA_TRAFOS = [
    "transforms_tef_model_zoo",
    "transforms_pyt_model_zoo"
]

transforms_pyt_model_zoo = transforms.Compose([
    transforms.Resize(256),
    transforms.CenterCrop(224),
    transforms.ToTensor(),
    transforms.Normalize(mean=[0.485, 0.456, 0.406],
                         std=[0.229, 0.224, 0.225]),
])
def transforms_tef_model_zoo(filename, label, image_size=224):

    img = cv.imread(filename.numpy().decode()).astype(np.float32)

    resize = 256
    if image_size > 224:
        resize = image_size

    # Resize
    height, width, _ = img.shape
    new_height = height * resize // min(img.shape[:2])
    new_width = width * resize // min(img.shape[:2])
    img = cv.resize(img, (new_width, new_height), interpolation=cv.INTER_CUBIC)

    # Crop
    height, width, _ = img.shape
    startx = width // 2 - (image_size // 2)
    starty = height // 2 - (image_size // 2)
    img = img[starty:starty + image_size, startx:startx + image_size]
    assert img.shape[0] == image_size and img.shape[1] == image_size, (img.shape, height, width)

    # BGR to RGB
    img = img[:, :, ::-1]

    return img, label

def model_transform_ImageNet_to_CIFAR100(original_model):
    if original_model.__class__.__name__ == 'ResNet':
        if hasattr(original_model, "fc"):
            classifier_in = original_model.fc.weight.shape[1]
            original_model.fc = nn.Linear(classifier_in, 100)
    elif 'MobileNet' in original_model.__class__.__name__ or\
            'EfficientNet' in original_model.__class__.__name__:
        idx = 1 if not 'V3' in original_model.__class__.__name__ else 3
        if hasattr(original_model, "classifier"):
            classifier_in = original_model.classifier[idx].weight.shape[1]
            original_model.classifier[idx] = nn.Linear(classifier_in, 100)
    return original_model

class ScaledConv2d(nn.Conv2d):
    def __init__(self, in_channels, out_channels, *args, **kwargs):
        super().__init__(in_channels, out_channels, *args, **kwargs)
        self.weight_scaling = nn.Parameter(torch.ones_like(torch.Tensor(out_channels, 1, 1, 1)))
        # self.reset_parameters()

    def reset_parameters(self):
        # The if condition is added so that the super call in init does not reset_parameters as well.
        if hasattr(self, 'weight_scaling'):
            nn.init.normal_(self.weight_scaling, 1, 0)#1e-5)
            super().reset_parameters()

    def forward(self, input):
        major, minor, _ = map(int, torch.__version__.split("."))
        if major > 1 or (major == 1 and minor > 7):
            return self._conv_forward(input, self.weight_scaling * self.weight, self.bias)
        else:
            return self._conv_forward(input, self.weight_scaling * self.weight)

class ScaledLinear(nn.Linear):
    def __init__(self, in_features, out_features, *args, **kwargs):
        super().__init__(in_features, out_features, *args, **kwargs)
        self.weight_scaling = nn.Parameter(torch.ones_like(torch.Tensor(out_features, 1)))
        # self.reset_parameters()

    def reset_parameters(self):
        # The if condition is added so that the super call in init does not reset_parameters as well.
            if hasattr(self, 'weight_scaling'):
                nn.init.normal_(self.weight_scaling, 1, 0)#1e-5)
                super().reset_parameters()

    def forward(self, input):
        return F.linear(input, self.weight_scaling * self.weight, self.bias)

class LSA:
    def __init__(self, original_model):
        self.mdl = copy.deepcopy(original_model)

    def update_conv2d(self, m, parent):
        lsa_update = ScaledConv2d(m[1].in_channels, m[1].out_channels, m[1].kernel_size, m[1].stride,
                                  m[1].padding, m[1].dilation, m[1].groups, None, m[1].padding_mode)
        lsa_update.weight, lsa_update.bias = m[1].weight, m[1].bias
        setattr(parent, m[0], lsa_update)

    def update_linear(self, m, parent):
        lsa_update = ScaledLinear(m[1].in_features, m[1].out_features)
        lsa_update.weight, lsa_update.bias = m[1].weight, m[1].bias
        setattr(parent, m[0], lsa_update)

    def add_lsa_params_recursive(self, module):
        for name, child in module.named_children():
            if isinstance(child, nn.Conv2d) and child.weight.requires_grad:
                self.update_conv2d((name, child), module)
            elif isinstance(child, nn.Linear) and child.weight.requires_grad:
                self.update_linear((name, child), module)
            elif len(list(child.children())) > 0:
                self.add_lsa_params_recursive(child)

    def add_lsa_params(self):
        self.add_lsa_params_recursive(self.mdl)
        return self.mdl


def split_datasets(testset, trainset, num_partitions: int, batch_size: int, num_workers: int, val_ratio: float = 0.1):

    # def seed_worker(worker_id):
    #     worker_seed = torch.initial_seed() % 2 ** 32
    #     np.random.seed(worker_seed)
    #     random.seed(worker_seed)
    #
    # g = torch.Generator()
    # g.manual_seed(0)

    # split trainset into `num_partitions` trainsets
    num_images = len(trainset) // num_partitions

    partition_len = [num_images] * num_partitions

    trainsets = random_split(
        trainset, partition_len, torch.Generator().manual_seed(2023)
    )

    # create dataloaders with train+val support
    trainloaders = []
    valloaders = []

    for trainset_ in trainsets:
        num_total = len(trainset_)
        num_val = int(val_ratio * num_total)
        num_train = num_total - num_val

        for_train, for_val = random_split(
            trainset_, [num_train, num_val], torch.Generator().manual_seed(909)
        )

        trainloaders.append(
            DataLoader(for_train, batch_size=batch_size, shuffle=True, num_workers=num_workers,)
                       # worker_init_fn=seed_worker)#, generator=g)
        )
        if val_ratio == 0:
            valloaders.append(DataLoader(testset, batch_size=batch_size))
        else:
            valloaders.append(
            DataLoader(for_val, batch_size=batch_size, shuffle=False, num_workers=num_workers,)
                       # worker_init_fn=seed_worker)#, generator=g)
            )

    # create dataloader for the test set
    testloader = DataLoader(testset, batch_size=batch_size)

    return trainloaders, valloaders, testloader



'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or
other Intellectual Property Rights other than the copyrights concerning
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''
import os.path
import random
import torch
import numpy as np
from collections import OrderedDict
import flwr as fl
from typing import Dict, Tuple
from flwr.common import (Scalar, ndarrays_to_parameters)

def model_diff(state_dict_a, state_dict_b): # new, old
    state_dict_diff = OrderedDict()
    for param_name in state_dict_a:
        state_dict_diff[param_name] = state_dict_a[param_name] - state_dict_b[param_name]
    return state_dict_diff

def model_add(state_dict_a, state_dict_b): # full, partial
    state_dict_sum = OrderedDict()
    for param_name in state_dict_a:
        if param_name in state_dict_b:
            state_dict_sum[param_name] = state_dict_a[param_name] + state_dict_b[param_name]
        else:
            state_dict_sum[param_name] = state_dict_a[param_name]
    return state_dict_sum
class NNClient(fl.client.NumPyClient):

    def __init__(self, context, trainloader, valloader, id, criterion, device, c_model, args, mdl_info,
                 encode_fn, decode_fn, train_fn) -> None:

        super().__init__()
        self.context = context
        self.trainloader = trainloader
        self.testloader = valloader
        self.model = c_model
        self.mdl_info = mdl_info
        self.criterion = criterion
        self.device = device
        try:
            self.optimizer = c_model.configure_optimizers(args.weight_decay, args.lr, (0.9, 0.95), device.type)
        except:
            self.optimizer = torch.optim.Adam(self.model.parameters(), lr=args.lr)
        self.accumulated_bs_sizes_per_round = 0
        self.id = id
        self.args = args
        self.encode_fn = encode_fn
        self.decode_fn = decode_fn
        self.train_fn = train_fn
        self.internal_states = {"approx_param_base": {"parameters": {},
                                                      "put_node_depth": {},
                                                      "device_id": self.id,
                                                      "parameter_id": {},
                                                      },
                                "comm_round": 0}
    def get_parameters(self, config
                       ) -> bytearray:
        """Extract model parameters and return them as a list of numpy arrays."""

        if self.args.compress_upstream:
            print("UP-STREAM compression:")
            param_dict = {k: np.float32(v.cpu().detach().numpy()) for k, v in self.model.state_dict().items()
                          if v.shape != torch.Size([])}

            if self.args.compress_differences: ## TODO: replace w/ checking mps_parent_signalling_enabled_flag
                if self.args.bnf:
                    param_dict = self.fold_bn(param_dict)
                param_dict = model_diff(param_dict, self.internal_states["prev_mdl"])

            if self.args.err_accumulation and "residuals" in self.internal_states:
                param_dict = model_add(param_dict, self.internal_states["residuals"])

            blkid_ptypes = {"block_identifier": self.mdl_info["block_identifier"],
                            "parameter_type": self.mdl_info["parameter_type"]} \
                             if self.args.bnf and not self.args.compress_differences else None

            params = [self.encode_fn(param_dict, vars(self.args), blkid_ptypes=blkid_ptypes,
                                     approx_param_base=self.internal_states["approx_param_base"],
                                     epoch=self.internal_states["comm_round"], device_id=self.id)]
            self.accumulated_bs_sizes_per_round += len(params[0])

            if self.args.err_accumulation:
                self.update_residual(compressed_update=self.decode_fn(params[0], approx_param_base=self.internal_states["approx_param_base"]),
                                     uncompressed_update=param_dict)
        else:
            params = [v.cpu().numpy() for _, v in self.model.state_dict().items() if v.shape != torch.Size([])]

        self.internal_states["comm_round"] += 1
        self.save_internal_states()

        return params

    def set_parameters(self, parameters):
        """Receive parameters and apply them to the local model."""

        if self.args.compress_downstream and len(parameters) == 1: # decoding
            parameters = ndarrays_to_parameters(parameters).tensors[0]
            decompressed_weights = self.decode_fn(bytearray(parameters[parameters.index(b'\n') + 1:]),
                                                  internal_states_path=f"{self.args.results}")

            if self.args.compress_differences and os.path.exists(f"{self.args.results}/client_ID{self.id}_internal_states.npz"):
                self.load_internal_states()
                decompressed_weights = model_add(self.internal_states["prev_mdl"], decompressed_weights)

            if self.args.bnf:
                decompressed_weights = self.unfold_bn(decompressed_weights)

            state_dict = OrderedDict({k: torch.tensor(v) for k, v in decompressed_weights.items()})

        else:
            state_dict = OrderedDict({k: torch.tensor(v) for k, v in zip(self.mdl_info["parameter_index"], parameters)})

        self.model.load_state_dict(state_dict, strict=True)

    def fit(self, parameters, config
            ) -> Tuple[bytearray, int, Dict[str, Scalar]]:
        """Train model received by the server (parameters) using the data.
           that belongs to this client. Then, send it back to the server.
        """

        if os.path.exists(f"{self.args.results}/client_ID{self.id}_internal_states.npz"):
            self.load_internal_states()


        self.set_parameters(parameters)
        self.internal_states["prev_mdl"] = {k: np.float32(v.cpu().detach().numpy())
                                            for k, v in self.model.state_dict().items()
                                            if v.shape != torch.Size([])}

        torch.manual_seed(808 + self.id + self.internal_states["comm_round"])
        np.random.seed(303 + self.id + self.internal_states["comm_round"])
        random.seed(909 + self.id + self.internal_states["comm_round"])

        train_res_dict = self.train_fn(self.model, optimizer=self.optimizer, criterion=self.criterion,
                                                    trainloader=self.trainloader, device=self.device, verbose=False,
                                                    args=self.args, round=self.internal_states["comm_round"])

        if self.args.bnf:
            self.internal_states["local_bn_params"] = {n: v.detach().cpu().numpy() for n, v in
                                                       self.model.state_dict().items() if
                                                       v.shape != torch.Size([]) and self.mdl_info["parameter_type"][n]
                                                       in ["bn.beta", "bn.gamma", "bn.mean", "bn.var"]}
            self.internal_states["prev_mdl"] = self.fold_bn(self.internal_states["prev_mdl"])

        print(f"Client ID: {self.id} {[str(k) + ': ' + str(v) for k, v in train_res_dict.items()]}")

        try:
            num_samples = len(self.trainloader)
        except:
            num_samples = self.trainloader.dataset.num_samples

        return self.get_parameters({}), num_samples, {**train_res_dict, "bs_size": self.accumulated_bs_sizes_per_round}

    def save_internal_states(self):
        np.savez(f"{self.args.results}/client_ID{self.id}_internal_states.npz", **self.internal_states)

    def load_internal_states(self):
        loaded_states = np.load(f"{self.args.results}/client_ID{self.id}_internal_states.npz",
                                allow_pickle=True)  # TODO get rid of allow_pickle
        self.internal_states = {k: loaded_states[k].item() for k in loaded_states.files}
        loaded_states.close()

    def update_residual(self, compressed_update, uncompressed_update):
        if not "residuals" in self.internal_states:
            self.internal_states["residuals"] = {}
        for k in uncompressed_update:
            ignored_param = False # k.endswith(".weight_scaling")
            if k in compressed_update and not ignored_param:
                self.internal_states["residuals"][k] = uncompressed_update[k] - compressed_update[k]
            else:
                self.internal_states["residuals"][k] = np.zeros_like(uncompressed_update[k]) \
                                                        if ignored_param else uncompressed_update[k]

    def fold_bn(self, param_dict):
        eps = 1e-3 if self.mdl_info['topology_storage_format'] == 4 else 1e-5

        for blk in self.mdl_info["mdl_blocks"]:
            block_access = self.mdl_info["mdl_blocks"][blk]
            if block_access["block_id"] is None:
                continue

            if "bn_mean" in block_access:
                bn_shape = param_dict[block_access["bn_mean"]].shape

                delta = block_access["bi"]
                if delta not in param_dict:
                    param_dict[delta] = np.zeros(bn_shape, dtype=np.float32)

                alpha = block_access["ls"]
                if alpha not in param_dict:
                    param_dict[alpha] = np.ones(bn_shape, dtype=np.float32)

                g = param_dict[block_access["bn_gamma"]] / np.sqrt(param_dict[block_access["bn_var"]] + eps)
                param_dict[alpha] *= g
                param_dict[delta] = (param_dict[delta] - param_dict[block_access["bn_mean"]]) * g + param_dict[block_access["bn_beta"]]

                del param_dict[block_access["bn_gamma"]], param_dict[block_access["bn_var"]],\
                    param_dict[block_access["bn_mean"]], param_dict[block_access["bn_beta"]]

        return param_dict

    def unfold_bn(self, folded_state_dict, momentum=0.3):
        eps = 1e-3 if self.mdl_info['topology_storage_format'] == 4 else 1e-5
        unfolded_state_dict = {}

        self.load_internal_states()
        local_last_round = self.internal_states["local_bn_params"]

        for k in self.mdl_info["parameter_index"]:

            if k in folded_state_dict: # non-BN params
                unfolded_state_dict[k] = folded_state_dict[k]

            else: # BN params
                block_id = self.mdl_info["block_identifier"][k]
                if block_id is None:
                    continue
                block_access = self.mdl_info["mdl_blocks"][block_id]

                if block_access["bn_beta"] not in unfolded_state_dict: # FedBNF
                    # BN beta update
                    unfolded_state_dict[block_access["bn_beta"]] = ((1 - momentum) * local_last_round[block_access["bn_beta"]]) +\
                                                                   (momentum * (folded_state_dict[block_access["bi"]] +
                                                                                (folded_state_dict[block_access["ls"]] *
                                                                                 local_last_round[block_access["bn_mean"]])))
                    # BN gamma update
                    unfolded_state_dict[block_access["bn_gamma"]] = ((1 - momentum) * local_last_round[block_access["bn_gamma"]]) +\
                                                                    (momentum * folded_state_dict[block_access["ls"]] *
                                                                     np.sqrt(local_last_round[block_access["bn_var"]] + eps))
                    # BN running stats
                    unfolded_state_dict[block_access["bn_mean"]] = local_last_round[block_access["bn_mean"]]
                    unfolded_state_dict[block_access["bn_var"]] = local_last_round[block_access["bn_var"]]

        return unfolded_state_dict



'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or
other Intellectual Property Rights other than the copyrights concerning
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''
import os
import glob
import random
import torch
import numpy as np

SEED_RANDOM = 909
SEED_TORCH = 808

def seed_worker(worker_id):
    seed = SEED_RANDOM + worker_id
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)

g = torch.Generator()
g.manual_seed(SEED_TORCH)


class PretokDatasetTelko(torch.utils.data.Dataset):
    """Loads pretokenized examples from disk and yields them as PyTorch tensors."""

    def __init__(self, max_seq_len, bin_path, split='train', shuffle=False):
        super().__init__()
        self.max_seq_len = max_seq_len
        self.bin_path = bin_path
        self.split = split
        self.shuffle = shuffle

        m = np.memmap(self.bin_path, dtype=np.uint8, mode="r")

        start_indices = np.where(m == 17)[0]
        if start_indices[0] != 0:
            start_indices = np.insert(start_indices, 0, 0)
        end_indices = np.append(start_indices[1:], len(m))
        self.sequences = [m[start:end] for start, end in zip(start_indices, end_indices)]

        self.num_samples = len(self.sequences)

        # Optional shuffling for training
        if self.shuffle:
            np.random.seed(42)  # or pass a seed argument
            np.random.shuffle(self.sequences)

    def __len__(self):
        return self.num_samples

    def __getitem__(self, idx):
        seq = self.sequences[idx]

        # Pad or truncate
        if len(seq) >= self.max_seq_len:
            padded = seq[:self.max_seq_len]
        else:
            pad_len = self.max_seq_len - len(seq)
            padding = np.zeros(pad_len, dtype=seq.dtype)
            padded = np.concatenate((seq, padding))

        chunk = torch.from_numpy(padded.astype(np.int32))
        x = chunk[:-1]
        y = chunk[1:]
        return x, y


class PretokDatasetTelkoTest(torch.utils.data.Dataset):
    """Loads pretokenized examples from a .npz file and returns them as PyTorch tensors."""

    def __init__(self, max_seq_len, bin_path, split='train', shuffle=False):
        super().__init__()
        self.max_seq_len = max_seq_len
        self.bin_path = bin_path
        self.split = split
        self.shuffle = shuffle

        m = np.memmap(self.bin_path, dtype=np.uint8, mode="r")

        start_indices = np.where(m == 17)[0]
        if start_indices[0] != 0:
            start_indices = np.insert(start_indices, 0, 0)
        end_indices = np.append(start_indices[1:], len(m))
        self.sequences = [m[start:end] for start, end in zip(start_indices, end_indices)]
        max_length = max(len(seq) for seq in self.sequences)
        print(f"Maximum sequence length: {max_length}")

        self.num_samples = len(self.sequences)

        # Optional shuffling for training
        if self.shuffle:
            np.random.seed(42)  # or pass a seed argument
            np.random.shuffle(self.sequences)

    def __len__(self):
        return self.num_samples

    def __getitem__(self, idx):
        seq = self.sequences[idx]

        # Pad or truncate
        if len(seq) >= self.max_seq_len:
            padded = seq[:self.max_seq_len]
        else:
            pad_len = self.max_seq_len - len(seq)
            padding = np.zeros(pad_len, dtype=seq.dtype)
            padded = np.concatenate((seq, padding))

        chunk = torch.from_numpy(padded.astype(np.int32))
        return chunk


def V2X(args, test_only=False, shuffle=False):

    if not test_only:

        print(f'searching: {os.path.join(args.dataset_path, "train")} and {os.path.join(args.dataset_path, "test")}')
        train_shard_filenames = sorted(glob.glob(os.path.join(os.path.join(args.dataset_path, 'train'), "*.bin")))

        test_shard_filename = sorted(glob.glob(os.path.join(os.path.join(args.dataset_path, 'test'), "*.bin")))

        if not len(train_shard_filenames) == args.num_clients:
            print(f"train_shard_filenames: {train_shard_filenames}")
            print(f"test_shard_filename: {test_shard_filename}")
            assert 0, f'number of pretokenized train files {len(train_shard_filenames)} does not match number of clients {args.num_clients}'

        train_loaders, val_loaders = [], []
        for shard in train_shard_filenames:
            ds = PretokDatasetTelko(args.max_seq_len, shard)
            dl = torch.utils.data.DataLoader(ds, batch_size=args.batch_size, pin_memory=not torch.backends.mps.is_available(), num_workers=args.workers,
                                             worker_init_fn=seed_worker, generator=g)
            train_loaders.append(dl)

            dlt = torch.utils.data.DataLoader(PretokDatasetTelko(args.max_seq_len, test_shard_filename[0], split='test'),
                                              batch_size=args.batch_size, pin_memory=not torch.backends.mps.is_available(), num_workers=args.workers,
                                              worker_init_fn=seed_worker, generator=g)
            val_loaders.append(dlt)

        test_loader = torch.utils.data.DataLoader(PretokDatasetTelko(args.max_seq_len, test_shard_filename[0], split='test'),
                                                  batch_size=args.batch_size, pin_memory=not torch.backends.mps.is_available(), num_workers=args.workers,
                                                  worker_init_fn=seed_worker, generator=g)
        return train_loaders, val_loaders, test_loader  ## Note: curently test loader and client val loaders are identical

    else:

        print(f'searching: {os.path.join(args.dataset_path, "test")}')
        test_shard_filename = sorted(glob.glob(os.path.join(os.path.join(args.dataset_path, 'test'), "*.bin")))

        test_loader = torch.utils.data.DataLoader(PretokDatasetTelkoTest(args.max_seq_len, test_shard_filename[0], split='test', shuffle=shuffle),
                                                  batch_size=args.batch_size, pin_memory= not torch.backends.mps.is_available(), num_workers=0,
                                                  worker_init_fn=seed_worker, generator=g)
        return test_loader


'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''

import copy, logging, os
import numpy as np
LOGGER = logging.getLogger()
from nncodec import nnc_core
from nncodec.framework.use_case_init import use_cases
from nncodec.framework.applications.utils import evaluation, transforms
import torch
from collections import OrderedDict


def is_pyt_model( model_object ):
    return isinstance( model_object, torch.nn.Module ) 

def __initialize_data_functions(
                 handler=None,
                 dataset_path=None,
                 batch_size=None,
                 num_workers=None,
                ):

    if dataset_path: 
        test_set, test_loader = handler.init_test(
            dataset_path,
            batch_size,
            num_workers
        )

        val_set, val_loader = handler.init_validation(
            dataset_path,
            batch_size,
            num_workers
        )

        train_loader = handler.init_training(
            dataset_path,
            batch_size,
            num_workers
        )
    else:
        test_set, test_loader = None, None
        val_set, val_loader   = None, None
        train_loader = None
    
    return test_set, test_loader, val_set, val_loader, train_loader

def create_NNC_model_instance_from_file(
                 model_path,
                 dataset_path=None,
                 lr=1e-4,
                 epochs=30,
                 max_batches=None,
                 batch_size=64,
                 num_workers=1,
                 model_struct=None,
                 lsa=False,
                 use_case=None
                ):

    PYTModel = PytorchModel()
    model_parameters, loaded_model_struct = PYTModel.load_model(model_path)
    if model_struct == None and loaded_model_struct != None:
        model_struct = loaded_model_struct

    if dataset_path and model_struct:
        PYTModelExecuter = create_model_executer(model_struct=model_struct,
                                                 dataset_path=dataset_path,
                                                 lr=lr,
                                                 epochs=epochs,
                                                 max_batches=max_batches,
                                                 batch_size=batch_size,
                                                 num_workers=num_workers,
                                                 lsa=lsa,
                                                 use_case=use_case
                                                 )
        if lsa:
            model_parameters = PYTModel.init_model_from_dict(PYTModelExecuter.model.state_dict())
    else:
        PYTModelExecuter = None

    return PYTModel, PYTModelExecuter, model_parameters


def create_NNC_model_instance_from_object(
                 model_object,
                 dataset_path=None,
                 lr=1e-4,
                 epochs=30,
                 max_batches=None,
                 batch_size=64,
                 num_workers=1,
                 model_struct=None,
                 lsa=False,
                 use_case=None
                ):

    PYTModel = PytorchModel()
    model_parameters, loaded_model_struct = PYTModel.init_model_from_model_object(model_object)
    if model_struct == None and loaded_model_struct != None:
        model_struct = loaded_model_struct

    if model_struct:
        PYTModelExecuter = create_model_executer(model_struct=model_struct,
                                                 dataset_path=dataset_path,
                                                 lr=lr,
                                                 epochs=epochs,
                                                 max_batches=max_batches,
                                                 batch_size=batch_size,
                                                 num_workers=num_workers,
                                                 lsa=lsa,
                                                 use_case=use_case
                                                 )
        if lsa:
            model_parameters = PYTModel.init_model_from_dict(PYTModelExecuter.model.state_dict())
    else:
        PYTModelExecuter = None

    return PYTModel, PYTModelExecuter, model_parameters


def create_model_executer(
                            model_struct,
                            dataset_path,
                            lr=1e-4,
                            epochs=30,
                            max_batches=None,
                            batch_size=64,
                            num_workers=1,
                            lsa=False,
                            use_case=None
                            ):
    
    assert model_struct != None, "model_struct must be specified in order to create a model_executer!"
    # assert dataset_path != None, "dataset_path must be specified in order to create a model_executer!"
    
    handler = use_cases['NNR_PYT'] if use_case == None else use_cases[f"{use_case}"]

    test_set, test_loader, val_set, val_loader, train_loader = __initialize_data_functions( handler=handler,
                                                                              dataset_path=dataset_path,
                                                                              batch_size=batch_size,
                                                                              num_workers=num_workers)

    # assert (test_set!=None and test_loader!= None) or ( val_set!= None and val_loader!= None ), "Any of the pairs test_set/test_loader or val_set/val_loader must be specified in order to use data driven optimizations methods!"
    PYTModelExecuter = ImageClassificationPytorchModelExecuter(handler,
                                                                   train_loader=train_loader,
                                                                   test_loader=test_loader,
                                                                   test_set=test_set,
                                                                   val_loader=val_loader,
                                                                   val_set=val_set,
                                                                   model_struct=model_struct,
                                                                   lsa=lsa,
                                                                   lr=lr,
                                                                   epochs=epochs,
                                                                   max_batches=max_batches)

    PYTModelExecuter.initialize_optimizer(lr=lr)

    return PYTModelExecuter


def save_to_pytorch_file( model_data, path ):
    model_dict = OrderedDict()
    for module_name in model_data:
        model_dict[module_name] = torch.tensor(model_data[module_name])
    torch.save(model_dict, path)

def np_to_torch(parameter_dict):
    return {name: torch.tensor(copy.deepcopy(parameter_dict[name])) for name in parameter_dict}

def torch_to_numpy(parameter_dict):
    return {name: copy.deepcopy(param).numpy() if param.get_device == -1 else
    copy.deepcopy(param).cpu().detach().numpy() for name, param in parameter_dict.items() if not "num_batches_tracked" in name}

def model_diff(state_dict_a, state_dict_b):
    state_dict_diff = OrderedDict()
    for param_name in state_dict_a:
        state_dict_diff[param_name] = state_dict_a[param_name] - state_dict_b[param_name]
    return state_dict_diff

def model_add(state_dict_a, state_dict_b):
    state_dict_sum = OrderedDict()
    for param_name in state_dict_a:
        if param_name in state_dict_b:
            state_dict_sum[param_name] = state_dict_a[param_name] + state_dict_b[param_name]
        else:
            state_dict_sum[param_name] = state_dict_a[param_name]
    return state_dict_sum
def get_model_file_with_parameters( parameters, model_struct ):

    new_model_struct = copy.deepcopy(model_struct)

    state_dict = OrderedDict()
    for param in parameters.keys():
        state_dict[param] = torch.tensor( parameters[param] )
        assert param in new_model_struct.state_dict(), "The provided model_strcut does not fit the parameter state dict decoded from the bitstream! Parameter '{}' not found in model_struct state dict!".format(param)

    new_model_struct.load_state_dict(state_dict)

    return new_model_struct


class PytorchModel(nnc_core.nnr_model.NNRModel):
    
    def __init__(self, model_dict=None):
        
        if model_dict and isinstance(model_dict, dict):
            self.init_model_from_dict(model_dict)
        else:
            self.__model_info = None

        if torch.backends.mps.is_available():
            self.device = torch.device("mps")
            torch.mps.manual_seed(909)
        else:
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            if torch.cuda.is_available():
                major, minor = map(int, torch.version.cuda.split("."))
                if major > 10 or (major == 10 and minor >= 2):
                    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"

    
    def load_model(self, 
                   model_path
                  ):
        
        model_file = torch.load(model_path, map_location=self.device) ##loads the state_dict
        model_struct = None
        
        try:
            model_parameter_dict = None

            # model state_dict
            if isinstance(model_file, OrderedDict):
                model_parameter_dict = model_file

            # checkpoint including state_dict
            elif isinstance(model_file, dict):
                for key in model_file.keys():
                    if isinstance(model_file[key], OrderedDict):
                        model_parameter_dict = model_file[key]
                        print("Loaded weights from state_dict '{}' from checkpoint elements {}".format(key,
                                                                                                       model_file.keys()))
                        break
                if not model_parameter_dict:
                    assert 0, "Checkpoint does not include a state_dict in {}".format(model_file.keys())

            # whole model (in general not recommended)
            elif isinstance(model_file, torch.nn.Module):
                model_parameter_dict = model_file.state_dict()
                model_struct = model_file

            # multi-GPU parallel trained models (torch.nn.DataParallel)
            if all(i[:7] == 'module.' for i in model_parameter_dict.keys()):
                print("Removing 'module.' prefixes from state_dict keys resulting from saving torch.nn.DataParallel "
                      "models not in the recommended way, that is torch.save(model.module.state_dict()")
                new_state_dict = OrderedDict()
                for n, t in model_parameter_dict.items():
                    name = n[7:]  # remove `module.`
                    new_state_dict[name] = t
                model_parameter_dict = new_state_dict

        except:
            raise SystemExit("Can't read model: {}".format(model_path))
        
        return self.init_model_from_dict(model_parameter_dict), model_struct ##intializes the model and the state dict and returns the PYTModel instance and parameters dict


    def init_model_from_model_object(   self,
                                        model_object,
                                    ):

        return self.init_model_from_dict(model_object.state_dict()), model_object


    @property
    def model_info(self):
        return self.__model_info

    def init_model_from_dict(self, pt_dict):

        if isinstance(pt_dict, dict):
            model_dict = pt_dict
        elif isinstance(pt_dict.state_dict(), dict):
            model_dict = pt_dict.state_dict()

        model_data = {'parameters': {}, 'reduction_method': 'baseline'}
        model_info = {'parameter_type': {}, 'parameter_dimensions': {}, 'parameter_index': {}, 'block_identifier': {}, 'original_size': {},
                      'topology_storage_format': nnc_core.nnr_model.TopologyStorageFormat.NNR_TPL_PYT,
                      'topology_compression_format': nnc_core.nnr_model.TopologyCompressionFormat.NNR_PT_RAW}

        # metadata only needed for MNASNet from PYT model zoo... further work: include into bitstream
        # self.metadata = getattr(model_dict, '_metadata', None)

        type_list_int = ['int8', 'int16', 'int32', 'uint8', 'uint16', 'uint32']
        type_list_1_bytes = ['int8', 'uint8']
        type_list_2_bytes = ['int16', 'uint16', 'float16']
        original_size = 0

        for i, module_name in enumerate(model_dict):
            if '.num_batches_tracked' in module_name:
                continue
            if model_dict[module_name].data.cpu().detach().numpy().dtype in type_list_1_bytes:
                original_size += model_dict[module_name].numel()
            elif model_dict[module_name].data.cpu().detach().numpy().dtype in type_list_2_bytes:
                original_size += model_dict[module_name].numel()*2
            else:
                original_size += model_dict[module_name].numel()*4
            model_data['parameters'][module_name] = np.int32(model_dict[module_name].data.cpu().detach().numpy()) if model_dict[module_name].data.cpu().detach().numpy().dtype in type_list_int else model_dict[module_name].data.cpu().detach().numpy()
            if '.weight_scaling' in module_name:
                model_data['parameters'][module_name] = model_data['parameters'][module_name].flatten()
            mdl_shape = model_data['parameters'][module_name].shape
            model_info['parameter_dimensions'][module_name] = mdl_shape
            if len(mdl_shape) == 0:  # scalar
                model_data['parameters'][module_name] = np.array([np.float32(model_data['parameters'][module_name])])
                model_info['parameter_dimensions'][module_name] = np.array([0]).shape
            model_info['parameter_index'][module_name] = i

            dims = len(mdl_shape)

            if dims > 1 and '.weight' in module_name:
                model_info['parameter_type'][module_name] = 'weight'
            elif dims > 1:
                model_info['parameter_type'][module_name] = 'weight'
            elif dims == 1:
                if '.bias' in module_name:
                    model_info['parameter_type'][module_name] = 'bias' 
                elif '.running_mean' in module_name:
                    model_info['parameter_type'][module_name] = 'bn.mean'
                elif '.running_var' in module_name:
                    model_info['parameter_type'][module_name] = 'bn.var'
                elif '.weight_scaling' in module_name:
                    model_info['parameter_type'][module_name] = 'weight.ls'
                elif 'gamma' in module_name:
                    model_info['parameter_type'][module_name] = 'bn.gamma'
                elif '.weight' in module_name:
                    model_info['parameter_type'][module_name] = "weight"
                else:
                    model_info['parameter_type'][module_name] = 'unspecified'
            else:
                model_info['parameter_type'][module_name] = 'unspecified'
            
        model_info["original_size"] = original_size

        self.__model_info = model_info

        return model_data["parameters"]

    def save_state_dict(self, path, model_data):
        model_dict = OrderedDict()
        for module_name in model_data:
            model_dict[module_name] = torch.tensor(model_data[module_name])
            if model_data[module_name].size == 1:
                model_dict[module_name] = torch.tensor(np.int64(model_data[module_name][0]))
        torch.save(model_dict, path)

    def get_torch_bn_info(self, module, bn_dict=None, prefix='', bn_param=False):
        if bn_dict is None:
            bn_dict = OrderedDict()
        for name, param in module._parameters.items():
            if param is not None:
                bn_dict[prefix + name] = bn_param
        for name, buf in module._buffers.items():
            if buf is not None and 'num_batches_tracked' not in name:
                bn_dict[prefix + name] = bn_param
        for n, m in module._modules.items():
            if m is not None:
                self.get_torch_bn_info(m, bn_dict, prefix + n + '.',
                                  bn_param=True if isinstance(m, torch.nn.BatchNorm2d) else False)
        return bn_dict
    def guess_block_id_and_param_type(self, model_parameters, bn_info=None):
        
        try:
            block_id_and_param_type = {"block_identifier" : {}, "parameter_type" : {}}
            block_dict = dict()
            blkNum = -1
            for param in model_parameters.keys():
                if "aux_classifier" in param: ## for DeepLabV3 demo
                    continue
                dims = len(model_parameters[param].shape)
                paramShape = model_parameters[param].shape
                splitted_param = param.split(".")
                param_end = splitted_param[-1]
                base_block_id  = ".".join(splitted_param[0:-1]+[""]) if len(splitted_param[0:-1]) != 0 else "genericBlk."


                if dims > 1 and ('kernel' in param_end or 'weight' in param_end):
                    paramType = 'weight'
                    blockId = base_block_id
                elif dims > 1:
                    paramType = 'weight'
                    blockId = base_block_id
                elif dims == 1:
                    if bn_info and not param.endswith(".weight_scaling") and bn_info[param]:
                        if 'weight' in param_end:
                            paramType = 'bn.gamma'
                            blockId = base_block_id
                        elif 'bias' in param_end:
                            paramType = 'bn.beta'
                            blockId = base_block_id
                        elif 'running_mean' in param_end:
                            paramType = 'bn.mean'
                            blockId = base_block_id
                        elif 'running_var' in param_end:
                            paramType = 'bn.var'
                            blockId = base_block_id
                    elif 'bias' in param_end or 'beta' in param_end:
                        paramType = 'bias'
                        blockId = base_block_id
                    elif 'running_mean' in param_end or 'moving_mean' in param_end:
                        paramType = 'bn.mean'
                        blockId = base_block_id
                    elif 'running_var' in param_end or 'moving_variance' in param_end:
                        paramType = 'bn.var'
                        blockId = base_block_id
                    elif 'weight_scaling' in param_end:
                        paramType = 'weight.ls'
                        blockId = base_block_id
                    elif 'gamma' in param_end:
                        paramType = 'bn.gamma'
                        blockId = base_block_id
                    elif 'weight' in param_end:
                        paramType = 'weight'
                        blockId = base_block_id
                    else:
                        paramType = 'unspecified'
                        blockId = None
                else:
                    paramType = 'unspecified'
                    blockId = None

                if blockId:
                    block_id = base_block_id + str(blkNum)
                    if block_id in block_dict.keys():
                        if any([a[1] == paramType for a in block_dict[block_id]]):
                            blkNum += 1
                        block_id = base_block_id + str(blkNum)
                        blockId = base_block_id + str(blkNum)
                    else:
                        blkNum += 1
                        block_id = base_block_id + str(blkNum)
                        blockId = base_block_id + str(blkNum)
                            
                    if block_id not in block_dict.keys():
                        block_dict[block_id] = []
                        
                    block_dict[block_id].append( [param, paramType, blockId, dims, paramShape] )
                else:
                    block_id_and_param_type["parameter_type"][param] = paramType
                    block_id_and_param_type["block_identifier"][param] = blockId
                            
            weight_block_list = []
            bn_block_list     = []

            for block_list in block_dict.values():
                if any(["bn." in a[1] for a in block_list]):
                    for i, val in enumerate(block_list):
                        par, parT, blkId, dims, _ = val
                        # if parT == 'weight' and dims == 1:
                        #     block_list[i][1] = "bn.gamma"
                        # if parT == 'bias':
                        #     block_list[i][1] = "bn.beta"
                    bn_block_list.append( block_list )
                else:
                    weight_block_list.append(block_list)

            for weight_block in weight_block_list:
                weight_shape = None
                weight_blkId = None
                for par, parT, blkId, dims, paramSh in weight_block:
                    block_id_and_param_type["parameter_type"][par] = parT
                    block_id_and_param_type["block_identifier"][par] = blkId
                    if parT == 'weight':
                        weight_shape = paramSh
                        weight_blkId = blkId
            
                if len(bn_block_list) != 0 and any([dim == bn_block_list[0][0][4][0] for dim in weight_shape]):
                    bn_block = bn_block_list.pop(0)
                    for par, parT, _, _, _ in bn_block:
                        block_id_and_param_type["parameter_type"][par] = parT
                        block_id_and_param_type["block_identifier"][par] = weight_blkId
            
            assert len(bn_block_list) == 0
                                
        except:
            print("INFO: Guessing of block_id_and_parameter_type failed! block_id_and_parameter_type has been set to 'None'!")
            block_id_and_param_type = None
                
        return block_id_and_param_type


class ImageClassificationPytorchModelExecuter(nnc_core.nnr_model.ModelExecute):

    def __init__(self, 
                 handler,
                 train_loader,
                 test_loader,
                 test_set,
                 val_loader,
                 val_set,
                 model_struct,
                 lsa,
                 max_batches=None,
                 epochs=5,
                 lr=1e-4,
                 ):

        if torch.backends.mps.is_available():
            self.device = torch.device("mps")
            torch.mps.manual_seed(909)
        else:
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            if torch.cuda.is_available():
                major, minor = map(int, torch.version.cuda.split("."))
                if major > 10 or (major == 10 and minor >= 2):
                    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
        
        torch.manual_seed(451)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        
        self.learning_rate = lr
        self.epochs = epochs
        self.max_batches = max_batches
        
        self.handle = handler
        if test_set:
            self.test_set = test_set
            self.test_loader = test_loader
        if val_set:
            self.val_set = val_set
            self.val_loader = val_loader
        if train_loader:
            self.train_loader = train_loader

        if model_struct and self.handle.model_transform != None and self.handle.model_transform.__name__ in transforms.MDL_TRAFOS:
            model_struct = self.handle.model_transform(model_struct)

        if model_struct:
            self.original_model = copy.deepcopy(model_struct)
            if lsa:
                lsa_gen = transforms.LSA(model_struct)
                model_struct = lsa_gen.add_lsa_params()
            self.model = model_struct
        else:
            self.original_model = None
            self.model = None
        self.rec_model = None

    def test_model(self,
                   parameters,
                   verbose=False
                   ):

        torch.set_num_threads(1)
        Model = copy.deepcopy(self.model)

        base_model_arch = Model.state_dict()
        model_dict = OrderedDict()
        for module_name in base_model_arch:
            if module_name in parameters:
                model_dict[module_name] = parameters[module_name] if torch.is_tensor(parameters[module_name]) else \
                    torch.tensor(parameters[module_name])
                if "weight_scaling" in module_name:
                   model_dict[module_name] = model_dict[module_name].reshape(base_model_arch[module_name].shape)

        # metadata only needed for MNASNet from PYT model zoo... further work: include into bitstream
        # if self.metadata is not None:
        #     model_dict._metadata = self.metadata

        for param in Model.state_dict().keys():
            if "num_batches_tracked" in param:
                continue
            assert param in parameters.keys(), "The provided model_strcut does not fit the parameter state dict decoded from the bitstream! Parameter '{}' not found in parameter dict!".format(param)
        
        Model.load_state_dict(model_dict)

        acc = self.handle.evaluate(
            Model,
            self.handle.criterion,
            self.test_loader,
            self.test_set,
            device=self.device,
            verbose=verbose
        )
        del Model
        return acc

    def eval_model(self,
                   parameters,
                   verbose=False
                   ):

        torch.set_num_threads(1)

        Model = copy.deepcopy(self.model)

        base_model_arch = Model.state_dict()
        model_dict = OrderedDict()
        for module_name in base_model_arch:
            if module_name in parameters:
                model_dict[module_name] = parameters[module_name] if torch.is_tensor(parameters[module_name]) else \
                    torch.tensor(parameters[module_name])
                if "weight_scaling" in module_name:
                   model_dict[module_name] = model_dict[module_name].reshape(base_model_arch[module_name].shape)

        # metadata only needed for MNASNet from PYT model zoo... further work: include into bitstream
        # if self.metadata is not None:
        #     model_dict._metadata = self.metadata

        Model.load_state_dict(model_dict)

        acc = self.handle.evaluate(
            Model,
            self.handle.criterion,
            self.val_loader,
            self.val_set,
            device=self.device,
            verbose=verbose
        )

        del Model
        return acc

    def tune_model(
            self,
            parameters,
            param_types,
            lsa_flag=False,
            ft_flag=False,
            verbose=False,
            wandb_logging=False
    ):
        if wandb_logging:
            import wandb
        torch.set_num_threads(1)
        verbose = 1 if (verbose & 1) else 0

        base_model_arch = self.model.state_dict()
        model_dict = OrderedDict()
        for module_name in base_model_arch:
            if module_name in parameters:
                model_dict[module_name] = parameters[module_name] if torch.is_tensor(parameters[module_name]) else \
                    torch.tensor(parameters[module_name])
                if "weight_scaling" in module_name:
                   model_dict[module_name] = model_dict[module_name].reshape(base_model_arch[module_name].shape)
        self.model.load_state_dict(model_dict)

        for param in parameters:
            parameters[param] = copy.deepcopy(self.model.state_dict()[param])

        tuning_params = []
        for name, param in self.model.named_parameters():
            if lsa_flag and ft_flag and param_types[name] in nnc_core.nnr_model.O_TYPES:
                param.requires_grad = True
                tuning_params.append(param)
            elif lsa_flag and param_types[name] == 'weight.ls':
                param.requires_grad = True
                tuning_params.append(param)
            elif ft_flag and param_types[name] != 'weight.ls' and param_types[name] in nnc_core.nnr_model.O_TYPES:
                param.requires_grad = True
                tuning_params.append(param)
            else:
                param.requires_grad = False

        self.tuning_optimizer = torch.optim.Adam(tuning_params, lr=self.learning_rate)

        perf = self.eval_model(parameters, verbose=verbose)
        best_loss, best_params = perf['mean_test_loss'], copy.deepcopy(parameters)
        if verbose:
            print(f'Validation accuracy (loss) before LSA and/or Fine Tuning: {perf["acc"]} ({perf["mean_test_loss"]})')
            print(f'Test performance (top1, top5, loss) before LSA and/or Fine Tuning: '
                  f'{self.test_model(parameters, verbose=verbose)}')
        for e in range(self.epochs):
            train_perf = self.handle.train(
                self.model,
                self.tuning_optimizer,
                self.handle.criterion,
                self.train_loader,
                device=self.device,
                verbose=verbose,
                freeze_batch_norm=True if lsa_flag and not ft_flag else False,
                max_batches=self.max_batches if self.max_batches else None
            )
            print(f'Epoch {e+1}: Train accuracy: {train_perf["acc"]}, Loss: {train_perf["mean_train_loss"]}')
            for param in parameters:
                parameters[param] = copy.deepcopy(self.model.state_dict()[param])
            perf = self.eval_model(parameters, verbose=verbose)
            if perf["mean_test_loss"] < best_loss and best_loss - perf["mean_test_loss"] > 1e-3:
                best_loss = perf["mean_test_loss"]
                best_params = copy.deepcopy(parameters)
            else:
                if verbose:
                    # print(f'Early Stopping due to model convergence or overfitting')
                    print(f'Epoch {e + 1}: Validation accuracy (loss) after Model Tuning: {perf["acc"]} ({perf["mean_test_loss"]})')
                # break
            if verbose:
                if lsa_flag and not ft_flag:
                    print(f'Epoch {e+1}: Validation accuracy (loss) after Model Tuning: {perf["acc"]} ({perf["mean_test_loss"]})')

            if wandb_logging:
                wandb.log({"LSA_val_acc": perf["acc"], "LSA_val_loss": perf["mean_test_loss"], "LSA_train_acc": train_perf["acc"], "LSA_train_loss": train_perf["mean_train_loss"]})
        if verbose or wandb_logging:
            test_perf = self.test_model(parameters, verbose=verbose)
            print(f'Test performance (top1, top5, loss) after LSA and/or Fine Tuning: {test_perf}')

        lsa_params, ft_params = {}, {}
        for name in best_params:
            if lsa_flag and param_types[name] == 'weight.ls':
                lsa_params[name] = best_params[name].cpu().numpy().flatten()
            elif ft_flag and param_types[name] != 'weight.ls' and param_types[name] in nnc_core.nnr_model.O_TYPES:
                ft_params[name] = best_params[name].cpu().numpy()

        if wandb_logging:
            wandb.log({"LSA_test_acc": test_perf["acc"], "LSA_test_loss": test_perf["mean_test_loss"]})
            wandb.log({"lsa_params": lsa_params})

        return (lsa_params, ft_params)


    def initialize_optimizer(self,
                             lr=1e-5,
                             mdl_params=None
                             ):
        if mdl_params:
            mdl_params_copy = copy.deepcopy(mdl_params)
            Model = self.model
            base_model_arch = Model.state_dict()
            model_dict = OrderedDict()
            for module_name in base_model_arch:
                if module_name in mdl_params_copy:
                    model_dict[module_name] = torch.tensor(mdl_params_copy[module_name])
                else:
                    model_dict[module_name] = base_model_arch[module_name]
            Model.load_state_dict(model_dict)

        if hasattr(self, "model"):
            params = [param for name, param in self.model.named_parameters()
                        if not '.weight_scaling' in name]
            self.optimizer = torch.optim.Adam(params, lr=lr)


    def has_eval(self):
        return True
    
    def has_test(self):
        return True
    
    def has_tune_ft(self):
        return True
    
    def has_tune_lsa(self):
        return True